#include <proto.h>

#ifndef GT_NO_SHM_LIB
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/Xlib.h>
#include <X11/extensions/XShm.h>
#endif

#include "gtimage.h"
#include "gterror.h"


/************************************************************************\
*
* GtImage class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

#define GTIF_SHM		0x01

#define GTSHM		0x01
#define GTAT		0x02
#define GTPMEM	0x04

#ifndef GT_NO_SHM_LIB
extern "C" int XShmGetEventBase(Display *);
#endif

/************************************************************************\
*
* Create and init gc.
*
\************************************************************************/
GtImage *GtMkImage(GtDisplay *d,GtTag *tgs)
{
	GtImage *obj;
	
	if(!(obj=new GtImage()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtImage object!\n");
		return(NULL);
	}
	if(!obj->Init(d,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init shared font.
*
\************************************************************************/
GtImage *GtMkImage(GtDisplay *d,ulong tag,...)
{
	va_list va;
	GtImage *obj;
	int p;
	
	va_start(va,tag);
	p=GtBeginVaTags();
	obj=GtMkImage(d,GtCopyVaTags(tag,&va));
	GtEndVaTags(p);
	va_end(va);
	return(obj);
}

/************************************************************************\
*
* Is shared memory images supported.
*
\************************************************************************/
sbool GtShmSupport(void)
{
#ifndef GT_NO_SHM_LIB
	return(TRUE);
#else
	return(FALSE);
#endif
}

/************************************************************************\
*
* Setup empty image.
*
\************************************************************************/
GtImage::GtImage(void)
{
	dsp=NULL;
	im=NULL;
	data=NULL;
	flags=0;
	comptype=0;
	send=FALSE;
	sinfo=NULL;
}

/************************************************************************\
*
* Destroy image.
*
\************************************************************************/
void GtImage::Destroy(void)
{
#ifndef GT_NO_SHM_LIB
	XShmSegmentInfo *shminfo;
	
	shminfo=(XShmSegmentInfo *)sinfo;
	if(flags & GTSHM)
	{
		if(shminfo)
		{
			if(flags & GTAT)
				XShmDetach(dsp->XDsp(),shminfo);
			if(im)
				XDestroyImage(im);
			if(shminfo->shmaddr)
				shmdt(shminfo->shmaddr);
			if(shminfo->shmid)
				shmctl(shminfo->shmid,IPC_RMID,0);
			free(shminfo);
		}
	}
	else
#endif
	{
		if(im)
			XDestroyImage(im);
/*
		if(data && (flags & GTPMEM))
			free(data);
*/
	}
	GtBase::Destroy();
}

/************************************************************************\
*
* Setup image.
*
\************************************************************************/
sbool GtImage::Init(GtDisplay *d,GtTag *tgs)
{
	GtTag *t;
	Visual *vis=NULL;
	unsigned int depth=0,w=1,h=1;
	int fmt=ZPixmap,off=0,bmpad=8,bpl=0;
#ifndef GT_NO_SHM_LIB
	XShmSegmentInfo *shminfo;
	int dmy1,dmy2;
	Bool pixm;
#endif

	dsp=d;
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Visual:
				vis=(Visual *)t->d.ptr;
				break;
			case GT_Depth:
				depth=t->d.val;
				break;
			case GT_XSize:
				w=t->d.val;
				break;
			case GT_YSize:
				h=t->d.val;
				break;
			case GT_Format:
				fmt=t->d.val;
				break;
			case GT_Offset:
				off=t->d.val;
				break;
			case GT_BitmapPad:
				bmpad=t->d.val;
				break;
			case GT_BytesPerLine:
				bpl=t->d.val;
				break;
			case GT_Data:
				data=(char *)t->d.ptr;
				break;
			case GT_Shared:
				if(t->d.val)
					flags|=GTSHM;
				else
					flags&=~GTSHM;
				break;
			case GT_SendComplete:
				if(t->d.val)
					send=TRUE;
				else
					send=FALSE;
				break;
		}
	}
	if(!dsp)
	{
		GtError(this,GTEC_NODISPLAY,"No GtDisplay!\n");
		return(FALSE);
	}
	if((w==0) || (h==0))
	{
		GtError(this,GTEC_MISC,"Illegal size of image: %d x %d\n",w,h);
		return(FALSE);
	}
	if(!vis)
		vis=dsp->XDefaultVisual();
	if(!depth)
		depth=dsp->XDefaultDepth();
	if(!bpl)
		bpl=w;
#ifndef GT_NO_SHM_LIB
	if((flags & GTSHM) && XShmQueryVersion(dsp->XDsp(),&dmy1,&dmy2,&pixm))
	{
		sinfo=malloc(sizeof(XShmSegmentInfo));
		if(!sinfo)
		{
			GtError(this,GTEC_NOMEMORY,"No memory shared segment info!\n");
			return(FALSE);
		}
		shminfo=(XShmSegmentInfo *)sinfo;
		shminfo->shmid=0;
		shminfo->shmaddr=NULL;
		shminfo->readOnly=False;
		if(!(im=XShmCreateImage(dsp->XDsp(),vis,depth,fmt,NULL,shminfo,w,h)))
		{
			GtError(this,GTEC_NOXRESOURCE,"No shared memory image!\n");
			return(FALSE);
		}
		if(!(shminfo->shmid=shmget(IPC_PRIVATE,im->bytes_per_line*im->height,
															IPC_CREAT|0777)))
		{
			GtError(this,GTEC_NOMEMORY,"No shared memory!\n");
			return(FALSE);
		}
		if(!(shminfo->shmaddr=(char *)shmat(shminfo->shmid,0,0)))
		{
			GtError(this,GTEC_MISC,"Unable to attach shared memory!\n");
			return(FALSE);
		}
		im->data=(char *)shminfo->shmaddr;
		if(XShmAttach(dsp->XDsp(),shminfo))
		{
			GtError(this,GTEC_MISC,"Unable to attach shared memory!\n");
			return(FALSE);
		}
		flags|=GTAT;
		comptype=XShmGetEventBase(dsp->XDsp())+ShmCompletion;
	}
	else
#endif
	{
		flags&=~GTSHM;
		if(!(im=XCreateImage(dsp->XDsp(),vis,depth,fmt,off,data,w,h,bmpad,bpl)))
		{
			GtError(this,GTEC_NOXRESOURCE,"No image!\n");
			return(FALSE);
		}
		if(!data && !(data=(char *)malloc(bpl*h)))
		{
			GtError(this,GTEC_NOMEMORY,"No memory for image!\n");
			return(FALSE);
		}
		im->data=data;
	}
	return(TRUE);
}

/************************************************************************\
*
* Find display tag and then setup normal.
*
\************************************************************************/
sbool GtImage::Init(GtTag *tgs)
{
	GtDisplay *d;
	
	d=(GtDisplay *)GtGetTagPtr(tgs,GT_Display,NULL);
	return(this->Init(d,tgs));
}

/************************************************************************\
*
* Set image attributes.
*
\************************************************************************/
sbool GtImage::Set(GtTag *tgs)
{
	GtTag *t;
	
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Data:
				im->data=(char *)t->d.ptr;
				break;
			case GT_SendComplete:
				if(t->d.val)
					send=TRUE;
				else
					send=FALSE;
				break;
		}
	}
	return(TRUE);
}

/************************************************************************\
*
* Get image attributes.
*
\************************************************************************/
void GtImage::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_Display:
				*ptr=(long)dsp;
				break;
			case GT_Depth:
				*ptr=im->depth;
				break;
			case GT_XSize:
				*ptr=im->width;
				break;
			case GT_YSize:
				*ptr=im->height;
				break;
			case GT_Format:
				*ptr=im->format;
				break;
			case GT_Offset:
				*ptr=im->xoffset;
				break;
			case GT_BitmapPad:
				*ptr=im->bitmap_pad;
				break;
			case GT_BytesPerLine:
				*ptr=im->bytes_per_line;
				break;
			case GT_Data:
				*ptr=(long)im->data;
				break;
			case GT_Shared:
				if(flags & GTSHM)
					*ptr=TRUE;
				else
					*ptr=FALSE;
				break;
			case GT_SendComplete:
				*ptr=send;
				break;
			case GT_XImage:
				*ptr=(long)im;
				break;
		}
	}
}

/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtImage::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtImage")==0)
		return(TRUE);
	else
		return(GtBase::IsSubClassOf(cn));
}

/************************************************************************\
*
* Copy image to drawable.
*
\************************************************************************/
void GtImage::PutImage(GtDrawable *dr,GtGC *gc,int sx,int sy,int dx,int dy,
												int xs,int ys)
{
	#ifndef GT_NO_SHM_LIB
	if(flags & GTIF_SHM)
		XShmPutImage(dsp->XDsp(),dr->XDr(),gc->XGC(),im,sx,sy,dx,dy,xs,ys,send);
	else
		XPutImage(dsp->XDsp(),dr->XDr(),gc->XGC(),im,sx,sy,dx,dy,xs,ys);
	#else
	XPutImage(dsp->XDsp(),dr->XDr(),gc->XGC(),im,sx,sy,dx,dy,xs,ys);
	#endif
}

/************************************************************************\
*
* Copy drawable to image.
*
\************************************************************************/
void GtImage::GetImage(GtDrawable *dr,int sx,int sy,int xs,int ys,ulong msk,
												int fmt,int dx,int dy)
{
	#ifndef GT_NO_SHM_LIB
	if(flags & GTIF_SHM)
		XShmGetImage(dsp->XDsp(),dr->XDr(),im,dx,dy,msk);
	else
		XGetSubImage(dsp->XDsp(),dr->XDr(),sx,sy,xs,ys,msk,fmt,im,dx,dy);
	#else
	XGetSubImage(dsp->XDsp(),dr->XDr(),sx,sy,xs,ys,msk,fmt,im,dx,dy);
	#endif
}
